# -*- coding: utf-8 -*-
"""add invadjust_reason

Revision ID: 0c91cf7d557b
Revises: a809caf23cf0
Create Date: 2018-01-05 18:14:18.544859

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '0c91cf7d557b'
down_revision = u'a809caf23cf0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # invadjust_reason
    op.create_table('invadjust_reason',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('code', sa.String(length=20), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=False),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('code', name=u'invadjust_reason_uq_code')
    )


def downgrade():

    # invadjust_reason
    op.drop_table('invadjust_reason')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
**Exceptions**

This module contains all core Rattail exception classes.
"""

from __future__ import unicode_literals, absolute_import

import six


class RattailError(Exception):
    """
    Base class for all Rattail exceptions.
    """


class ConfigurationError(RattailError):
    """
    Generic class for configuration errors.
    """


@six.python_2_unicode_compatible
class SQLAlchemyNotInstalled(RattailError):
    """
    Error raised when an operation is requested which requires SQLAlchemy
    (and/or related libraries) to be installed, but they are not available.
    """
    def __init__(self, error):
        self.error = error

    def __str__(self):
        return ("Hm, looks like SQLAlchemy may not be installed?  "
                "(Perhaps you should do: 'pip install rattail[db]'?)  "
                "Original error was: {}".format(self.error))


@six.python_2_unicode_compatible
class WindowsExtensionsNotInstalled(RattailError):
    """
    Error raised when an operation is requested which requires the "Python for
    Windows Extensions" to be instealled, but it is not available.
    """

    def __str__(self):
        return ("Cannot proceed because Python for Windows Extensions is not installed.  Please see "
                "https://rattailproject.org/moin/Installation/Windows/Python#Python_for_Windows_Extensions "
                "for more info.")


@six.python_2_unicode_compatible
class MailTemplateNotFound(ConfigurationError):

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return ("No message templates could be found for '{0}' emails.  Please "
                "create '{0}.(txt|html).mako' and place it/them in one of the "
                "configured template folders.".format(self.key))


@six.python_2_unicode_compatible
class SenderNotFound(ConfigurationError):

    def __init__(self, key):
        self.key = key

    def __str__(self):
        return ("No email sender (From: address) found in config.  Please set '{}.from' "
                "(or 'default.from') in the [rattail.mail] section.".format(self.key))


@six.python_2_unicode_compatible
class RecipientsNotFound(ConfigurationError):
    
    def __init__(self, key):
        self.key = key

    def __str__(self):
        return ("No recipients found in config for '{0}' emails.  Please set '{0}.to' "
                "(or 'default.to') in the [rattail.mail] section.".format(self.key))


class FileOperationError(RattailError):
    """
    Generic exception for file operation failures.
    """


@six.python_2_unicode_compatible
class PathNotFound(FileOperationError):
    """
    Raised when "path not found" errors are encountered within the
    :func:`rattail.files.locking_copy_test()` function.  The purpose of this is
    to normalize these errors to a single type, since the file monitor retry
    mechanism will fail if two distinct exceptions are encountered during its
    processing attempts.
    """
    def __init__(self, original_error):
        self.original_error = original_error

    def __str__(self):
        return '{}: {}'.format(
            self.original_error.__class__.__name__,
            self.original_error)


@six.python_2_unicode_compatible
class BatchAlreadyExecuted(RattailError):

    def __init__(self, batch):
        self.batch = batch

    def __str__(self):
        return "Batch is already executed: {}".format(self.batch)


class LabelPrintingError(Exception):

    pass


class PalmError(RattailError):
    """
    Base class for all errors relating to the Palm OS application interface.
    """


@six.python_2_unicode_compatible
class PalmClassicDatabaseTypelibNotFound(PalmError):

    def __str__(self):
        return ("The Python module for the Palm Classic Database type library "
                "could not be generated.  (Is the HotSync Manager software "
                "installed?)")


@six.python_2_unicode_compatible
class PalmConduitManagerNotFound(PalmError):

    def __str__(self):
        return ("The Palm Desktop Conduit Manager could not be instantiated.  "
                "(Is the HotSync Manager software installed?)")


@six.python_2_unicode_compatible
class PalmConduitAlreadyRegistered(PalmError):

    def __str__(self):
        return "The Rattail Palm conduit is already registered."


@six.python_2_unicode_compatible
class PalmConduitNotRegistered(PalmError):

    def __str__(self):
        return "The Rattail Palm conduit is not registered."


class StopProcessing(RattailError):
    """
    Simple exception to indicate action processing should stop.  This
    is probably only useful for tests.
    """



# -*- coding: utf-8; -*-
"""add item_id for product batches

Revision ID: fd935205655d
Revises: 69d924015613
Create Date: 2017-07-18 23:39:02.717734

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'fd935205655d'
down_revision = u'69d924015613'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch*
    op.add_column('batch_handheld_row', sa.Column('item_id', sa.String(length=20), nullable=True))
    op.add_column('batch_inventory_row', sa.Column('item_id', sa.String(length=20), nullable=True))
    op.add_column('batch_pricing_row', sa.Column('item_id', sa.String(length=20), nullable=True))
    op.add_column('label_batch_row', sa.Column('item_id', sa.String(length=20), nullable=True))
    op.add_column('vendor_invoice_row', sa.Column('item_id', sa.String(length=20), nullable=True))


def downgrade():

    # batch*
    op.drop_column('vendor_invoice_row', 'item_id')
    op.drop_column('label_batch_row', 'item_id')
    op.drop_column('batch_pricing_row', 'item_id')
    op.drop_column('batch_inventory_row', 'item_id')
    op.drop_column('batch_handheld_row', 'item_id')



# -*- coding: utf-8; -*-
"""add purchase_batch.truck_dump_status

Revision ID: 03be1584f923
Revises: 40a4855cb22d
Create Date: 2019-03-01 10:50:37.053705

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '03be1584f923'
down_revision = '40a4855cb22d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


BATCH_STATUS_TRUCKDUMP_UNCLAIMED        = 3
BATCH_STATUS_TRUCKDUMP_CLAIMED          = 4

ROW_STATUS_TRUCKDUMP_UNCLAIMED          = 7
ROW_STATUS_TRUCKDUMP_CLAIMED            = 8
ROW_STATUS_TRUCKDUMP_OVERCLAIMED        = 9
ROW_STATUS_TRUCKDUMP_PARTCLAIMED        = 11


def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('truck_dump_status', sa.Integer(), nullable=True))
    batch = sa.sql.table('purchase_batch',
                         sa.sql.column('status_code'),
                         sa.sql.column('truck_dump_status'))
    op.execute(batch.update()\
               .where(batch.c.status_code == BATCH_STATUS_TRUCKDUMP_UNCLAIMED)\
               .values({'truck_dump_status': BATCH_STATUS_TRUCKDUMP_UNCLAIMED}))
    op.execute(batch.update()\
               .where(batch.c.status_code == BATCH_STATUS_TRUCKDUMP_CLAIMED)\
               .values({'truck_dump_status': BATCH_STATUS_TRUCKDUMP_CLAIMED}))

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('truck_dump_status', sa.Integer(), nullable=True))
    row = sa.sql.table('purchase_batch_row',
                       sa.sql.column('status_code'),
                       sa.sql.column('truck_dump_status'))
    op.execute(row.update()\
               .where(row.c.status_code == ROW_STATUS_TRUCKDUMP_UNCLAIMED)\
               .values({'truck_dump_status': ROW_STATUS_TRUCKDUMP_UNCLAIMED}))
    op.execute(row.update()\
               .where(row.c.status_code == ROW_STATUS_TRUCKDUMP_CLAIMED)\
               .values({'truck_dump_status': ROW_STATUS_TRUCKDUMP_CLAIMED}))
    op.execute(row.update()\
               .where(row.c.status_code == ROW_STATUS_TRUCKDUMP_OVERCLAIMED)\
               .values({'truck_dump_status': ROW_STATUS_TRUCKDUMP_OVERCLAIMED}))
    op.execute(row.update()\
               .where(row.c.status_code == ROW_STATUS_TRUCKDUMP_PARTCLAIMED)\
               .values({'truck_dump_status': ROW_STATUS_TRUCKDUMP_PARTCLAIMED}))


def downgrade():

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'truck_dump_status')

    # purchase_batch
    op.drop_column('purchase_batch', 'truck_dump_status')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Users API
"""

from __future__ import unicode_literals, absolute_import

import six
from sqlalchemy import orm

from rattail.db import model


def make_username(person):
    """
    Make up a new username for the given person.  Takes the often-used approach
    of first initial plus last name, e.g. 'ledgar' for Lance Edgar.  Appends an
    autoincrement number if necessary, until a unique name is found.
    """
    session = orm.object_session(person)
    users = session.query(model.User)

    basename = (person.last_name or '').strip().lower()
    if basename:
        if (person.first_name or '').strip():
            basename = '{}{}'.format((person.first_name or '').strip().lower()[0],
                                     basename)
    else:
        basename = (person.first_name or '').strip().lower()

    if basename:
        i = 0
        username = basename
    else:
        i = 1
        username = six.text_type(i)
    while users.filter_by(username=username).count():
        i += 1
        username = '{}{}'.format(basename, i)

    return username



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for IFPS
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa

from rattail.db.model import Base, uuid_column


@six.python_2_unicode_compatible
class IFPS_PLU(Base):
    """
    IFPS PLU codes
    """
    __tablename__ = 'ifps_plu'
    __table_args__ = (
        sa.UniqueConstraint('plu', name='ifps_plu_uq_plu'),
    )
    __versioned__ = {}
    model_title = "IFPS PLU Code"

    uuid = uuid_column()

    plu = sa.Column(sa.Integer(), nullable=False)
    category = sa.Column(sa.String(length=100), nullable=True)
    commodity = sa.Column(sa.String(length=100), nullable=True)
    variety = sa.Column(sa.Text(), nullable=True)
    size = sa.Column(sa.String(length=100), nullable=True)
    measurements_north_america = sa.Column(sa.String(length=255), nullable=True)
    measurements_rest_of_world = sa.Column(sa.String(length=255), nullable=True)
    restrictions_notes = sa.Column(sa.Text(), nullable=True)
    botanical_name = sa.Column(sa.String(length=100), nullable=True)
    aka = sa.Column(sa.Text(), nullable=True)
    revision_date = sa.Column(sa.DateTime(), nullable=True)
    date_added = sa.Column(sa.DateTime(), nullable=True)
    gpc = sa.Column(sa.String(length=100), nullable=True)
    image = sa.Column(sa.String(length=255), nullable=True)
    image_source = sa.Column(sa.String(length=100), nullable=True)

    def __str__(self):
        return str(self.plu)



# -*- coding: utf-8; -*-
"""add vendor catalog unit cost diff

Revision ID: 40c0c107ee6d
Revises: 5ff1dddc0a30
Create Date: 2019-10-09 13:50:57.673800

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '40c0c107ee6d'
down_revision = '5ff1dddc0a30'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor_catalog_row
    op.add_column('vendor_catalog_row', sa.Column('unit_cost_diff_percent', sa.Numeric(precision=10, scale=5), nullable=True))


def downgrade():

    # vendor_catalog_row
    op.drop_column('vendor_catalog_row', 'unit_cost_diff_percent')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Generic web API client base class
"""

import requests


class GenericWebAPI:
    """
    Generic base class for web API clients.
    """

    def __init__(
            self,
            config,
            base_url,
            max_retries=None,
            verify=None,
            **kwargs):

        self.config = config
        self.app = self.config.get_app()
        self.model = self.app.model

        self.session = requests.Session()

        self.base_url = base_url
        if self.base_url:
            self.base_url = self.base_url.rstrip('/')

        self.max_retries = max_retries
        if self.max_retries is not None:
            adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
            self.session.mount(self.base_url, adapter)

        if verify is not None:
            self.session.verify = verify

    def _request(self, request_method, api_method, params=None, data=None):
        """
        Perform a request for the given API method, and return the response.
        """
        api_method = api_method.lstrip('/')
        url = f'{self.base_url}/{api_method}'
        params = params or {}

        if request_method == 'GET':
            response = self.session.get(url, params=params)

        elif request_method == 'POST':
            kwargs = {'params': params}
            if data:
                kwargs['data'] = json.dumps(data)
            response = self.session.post(url, **kwargs)

        else:
            raise NotImplementedError(f"unknown request method: {request_method}")

        response.raise_for_status()
        return response

    def get(self, api_method, params=None):
        """
        Perform a GET request for the given API method, and return the response.
        """
        return self._request('GET', api_method, params=params)

    def post(self, api_method, params=None):
        """
        Perform a POST request for the given API method, and return the response.
        """
        return self._request('POST', api_method, params=params)

    def patch(self, api_method, params=None):
        """
        Perform a PATCH request for the given API method, and return the response.
        """
        return self._request('PATCH', api_method, params=params)

    def delete(self, api_method, params=None):
        """
        Perform a DELETE request for the given API method, and return the response.
        """
        return self._request('DELETE', api_method, params=params)



# -*- coding: utf-8; -*-
"""add tailbone_page_help

Revision ID: 31b0b868d6f0
Revises: 4594843ad852
Create Date: 2022-12-24 14:04:59.538930

"""

# revision identifiers, used by Alembic.
revision = '31b0b868d6f0'
down_revision = '4594843ad852'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # tailbone_page_help
    op.create_table('tailbone_page_help',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('route_prefix', sa.String(length=254), nullable=False),
                    sa.Column('help_url', sa.String(length=254), nullable=True),
                    sa.Column('markdown_text', sa.Text(), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('route_prefix', name='tailbone_page_help_uq_route_prefix')
    )
    op.create_table('tailbone_page_help_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('route_prefix', sa.String(length=254), autoincrement=False, nullable=True),
                    sa.Column('help_url', sa.String(length=254), autoincrement=False, nullable=True),
                    sa.Column('markdown_text', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_tailbone_page_help_version_end_transaction_id'), 'tailbone_page_help_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_tailbone_page_help_version_operation_type'), 'tailbone_page_help_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_tailbone_page_help_version_transaction_id'), 'tailbone_page_help_version', ['transaction_id'], unique=False)

    # tailbone_field_info
    op.create_table('tailbone_field_info',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('route_prefix', sa.String(length=254), nullable=False),
                    sa.Column('field_name', sa.String(length=100), nullable=False),
                    sa.Column('markdown_text', sa.Text(), nullable=True),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('route_prefix', 'field_name', name='tailbone_field_info_uq_field')
    )
    op.create_table('tailbone_field_info_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('route_prefix', sa.String(length=254), autoincrement=False, nullable=True),
                    sa.Column('field_name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('markdown_text', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_tailbone_field_info_version_end_transaction_id'), 'tailbone_field_info_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_tailbone_field_info_version_operation_type'), 'tailbone_field_info_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_tailbone_field_info_version_transaction_id'), 'tailbone_field_info_version', ['transaction_id'], unique=False)


def downgrade():

    # tailbone_field_info
    op.drop_index(op.f('ix_tailbone_field_info_version_transaction_id'), table_name='tailbone_field_info_version')
    op.drop_index(op.f('ix_tailbone_field_info_version_operation_type'), table_name='tailbone_field_info_version')
    op.drop_index(op.f('ix_tailbone_field_info_version_end_transaction_id'), table_name='tailbone_field_info_version')
    op.drop_table('tailbone_field_info_version')
    op.drop_table('tailbone_field_info')

    # tailbone_page_help
    op.drop_index(op.f('ix_tailbone_page_help_version_transaction_id'), table_name='tailbone_page_help_version')
    op.drop_index(op.f('ix_tailbone_page_help_version_operation_type'), table_name='tailbone_page_help_version')
    op.drop_index(op.f('ix_tailbone_page_help_version_end_transaction_id'), table_name='tailbone_page_help_version')
    op.drop_table('tailbone_page_help_version')
    op.drop_table('tailbone_page_help')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Telemetry Commands
"""

import logging
import pprint

from rattail.commands import Subcommand


log = logging.getLogger(__name__)


class Telemetry(Subcommand):
    """
    Submit telemetry data to a server
    """
    name = 'telemetry'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--profile', '-p', metavar='KEY',
                            help="Profile (type) of telemetry data to collect.  "
                            "This also determines where/how data is submitted.  "
                            "If not specified, default logic is assumed.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through all the motions but do not submit "
                            "the data to server.")

    def run(self, args):
        self.handler = self.app.get_telemetry_handler()

        data = self.handler.collect_all_data(profile=args.profile)
        log.info("data collected okay: %s", ', '.join(sorted(data)))
        log.debug("%s", data)

        if self.verbose:
            print("COLLECTED DATA:")
            pprint.pprint(data)

        if args.dry_run:
            log.info("dry run, so will not submit data to server")
        else:
            self.handler.submit_all_data(data, profile=args.profile)
            log.info("data submitted okay")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
People Handler

See also :doc:`rattail-manual:base/handlers/other/people`.
"""

import warnings

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.app import GenericHandler, MergeMixin
from rattail.time import make_utc


class PeopleHandler(GenericHandler, MergeMixin):
    """
    Base class and default implementation for people handlers.
    """

    def get_person(self, obj, **kwargs):
        """
        Retrieve the Person related to the given object.

        This is a rather fundamental method, in that it is called by
        several other methods, both within this handler as well as
        others.  There is even a shortcut to it, accessible via
        :meth:`rattail.app.AppHandler.get_person()`.

        Its purpose is to navigate relationships as needed, to get at
        the "default" person associated with the object.  Depending on
        how the app tracks relationships, this logic may need to vary.
        """
        model = self.model

        if isinstance(obj, model.Person):
            return obj

        elif isinstance(obj, model.User):
            if obj.person:
                return obj.person

        elif isinstance(obj, model.Employee):
            return obj.person

        elif isinstance(obj, model.Customer):
            if obj.account_holder:
                return obj.account_holder
            if obj.shoppers:
                return obj.shoppers[0].person
            # legacy fallback
            if obj.people:
                return obj.people[0]

        elif isinstance(obj, model.Member):
            if obj.person:
                return obj.person

    def should_expose_quickie_search(self):
        return self.config.getbool('rattail',
                                   'people.expose_quickie_search',
                                   default=False)

    def get_quickie_search_placeholder(self):
        return self.app.get_customer_key_label()

    def quickie_lookup(self, entry, session):
        """
        Attempt to locate a person based on the given entry.  By
        default this will do a lookup based on the configured Customer
        key field.  Override as needed.
        """
        model = self.model
        field = self.app.get_customer_key_field()

        # validate/coerce data type
        # TOOD: this is a hack! should inspect field type etc.
        if field == 'number':
            if not entry.isdigit():
                return
            entry = int(entry)

        try:
            customer = session.query(model.Customer)\
                              .filter(getattr(model.Customer, field) == entry)\
                              .one()
        except orm.exc.NoResultFound:
            pass
        else:
            return self.app.get_person(customer)

    def normalize_full_name(self, first, last, **kwargs):
        """
        Normalize a "full" name based on the given first and last
        names.  Tries to be smart about collapsing whitespace etc.

        :param first: First name.
        :param last: Last name.
        :returns: First and last name combined.
        """
        from rattail.db.util import normalize_full_name
        return normalize_full_name(first, last)

    def make_person(self, **kwargs):
        """
        Make and return a new Person instance.
        """
        model = self.model
        person = model.Person()

        if 'first_name' in kwargs:
            person.first_name = kwargs.pop('first_name')
        if 'middle_name' in kwargs:
            person.middle_name = kwargs.pop('middle_name')
        if 'last_name' in kwargs:
            person.last_name = kwargs.pop('last_name')

        if 'display_name' in kwargs:
            person.display_name = kwargs.pop('display_name')
        else:
            person.display_name = self.normalize_full_name(
                person.first_name, person.last_name)

        for key, value in kwargs.items():
            if hasattr(person, key):
                setattr(person, key, value)

        return person

    def update_names(self, person, **kwargs):
        """
        Update name(s) for the given person.

        :param person: Reference to a ``Person`` record.
        :param first: First name for the person.
        :param middle: Middle name for the person.
        :param last: Last name for the person.
        :param full: Full (display) name for the person.
        """
        if 'first' in kwargs:
            person.first_name = kwargs['first']

        if 'middle' in kwargs:
            person.middle_name = kwargs['middle']

        if 'last' in kwargs:
            person.last_name = kwargs['last']

        if 'full' in kwargs:
            if kwargs['full']:
                person.display_name = kwargs['full']
            else:
                person.display_name = self.normalize_full_name(
                    person.first_name, person.last_name)
        elif 'first' in kwargs and 'last' in kwargs:
            person.display_name = self.normalize_full_name(
                person.first_name, person.last_name)

    def add_phone(self, person, number, type='Home', preferred=False, **kwargs):
        """
        Add a phone record for the person.

        :param person: Reference to a ``Person`` record.
        :param number: Actual phone number to add.
        :param type: Type of phone number to add.
        :param preferred: Boolean indicating that this should be the
           new "preferred" number for the person.
        """
        reason = self.app.phone_number_is_invalid(number)
        if reason:
            raise ValueError("Phone number is not valid: {}".format(reason))

        phone = person.add_phone(number=self.app.format_phone_number(number),
                                 type=type)
        if preferred:
            person.set_primary_phone(phone)

        return phone

    def update_phone(self, person, phone, **kwargs):
        """
        Update a phone record for the person.

        :param person: Reference to a ``Person`` record.
        :param phone: Reference to the ``PersonPhoneNumber`` record to update.
        :param number: Actual phone number.
        :param type: Type of phone number.
        :param preferred: Boolean indicating that this should be the
           new "preferred" number for the person.
        """
        if phone not in person.phones:
            raise ValueError("Phone does not belong to this person.")

        if 'number' in kwargs:
            reason = self.app.phone_number_is_invalid(kwargs['number'])
            if reason:
                raise ValueError("Phone number is not valid: {}".format(reason))
            phone.number = self.app.format_phone_number(kwargs['number'])

        if 'type' in kwargs:
            phone.type = kwargs['type']

        if 'preferred' in kwargs:
            if kwargs['preferred']:
                person.set_primary_phone(phone)
            else: # should *not* prefer this one
                if phone.preferred and len(person.phones) > 1:
                    # make 2nd phone the new 1st
                    person.set_primary_phone(person.phones[1])

        return phone

    def add_email(self, person, address, type='Home',
                  invalid=False, preferred=False, **kwargs):
        """
        Add a email record for the person.

        :param person: Reference to a ``Person`` record.
        :param address: Actual email address to add.
        :param type: Type of email address to add.
        :param invalid: Boolean indicating the address is known to
           *not* be valid.
        :param preferred: Boolean indicating that this should (or not)
           be the "preferred" email address for the person.
        """
        email = person.add_email(address=address,
                                 type=type,
                                 invalid=invalid)
        if preferred:
            person.set_primary_email(email)

        return email

    def update_email(self, person, email, **kwargs):
        """
        Update a email record for the person.

        :param person: Reference to a ``Person`` record.
        :param email: Reference to the ``PersonEmailAddress`` record to update.
        :param address: Actual email address.
        :param type: Type of email address.
        :param invalid: Boolean indicating the address is known to
           *not* be valid.
        :param preferred: Boolean indicating that this should (or not)
           be the "preferred" address for the person.
        """
        if email not in person.emails:
            raise ValueError("Email does not belong to this person.")

        address_changed = False
        if 'address' in kwargs:
            address = kwargs['address']
            if self.should_force_email_to_lower_case():
                address = address.lower()
            if email.address != address:
                email.address = address
                address_changed = True

        if 'type' in kwargs:
            email.type = kwargs['type']

        if address_changed:
            if email.invalid:
                email.invalid = False
        elif 'invalid' in kwargs:
            email.invalid = kwargs['invalid']

        if 'preferred' in kwargs:
            if kwargs['preferred']:
                person.set_primary_email(email)
            else: # should *not* prefer this one
                if email.preferred and len(person.emails) > 1:
                    # make 2nd email the new 1st
                    person.set_primary_email(person.emails[1])

        return email

    def should_force_email_to_lower_case(self):
        return self.config.getbool('rattail',
                                   'people.should_force_email_to_lower_case',
                                   default=False)

    def add_address(self, person, type=None,
                    street=None, street2=None,
                    city=None, state=None, zipcode=None,
                    invalid=False, preferred=False, **kwargs):
        """
        Add a physical/mailing address record for the person.

        :param person: Reference to a ``Person`` record.
        :param type: Type of address to add.
        :param street: Street (line 1) for the address.
        :param street2: Street (line 2) for the address.
        :param city: City for the address.
        :param state: State for the address.
        :param zipcode: Zipcode for the address.
        :param invalid: Boolean indicating the address is known to
           *not* be valid.
        :param preferred: Boolean indicating that this should (or not)
           be the "preferred" address for the person.
        """
        address = person.add_address(type=type,
                                     street=street,
                                     street2=street2,
                                     city=city,
                                     state=state,
                                     zipcode=zipcode,
                                     invalid=invalid,
                                     flush=False)
        if preferred:
            person.set_primary_address(address)

        return address

    def ensure_address(self, person, **kwargs):
        """
        Returns the default address record associated with the given
        person, creating it first if necessary.
        """
        address = person.first_address()
        if not address:
            address = self.add_address(person, **kwargs)
        return address

    def update_address(self, person, address, **kwargs):
        """
        Update the given address with the given data.
        """
        if 'type' in kwargs:
            address.type = kwargs['type']

        if 'street' in kwargs:
            address.street = kwargs['street']

        if 'street2' in kwargs:
            address.street2 = kwargs['street2']

        if 'city' in kwargs:
            address.city = kwargs['city']

        if 'state' in kwargs:
            address.state = kwargs['state']

        if 'zipcode' in kwargs:
            address.zipcode = kwargs['zipcode']

        if 'invalid' in kwargs:
            self.mark_address_invalid(person, address, kwargs['invalid'])

    def address_is_invalid(self, person, address, **kwargs):
        """
        Returns a boolean indicating if the given person's address is
        invalid.

        :param person: Reference to a person.
        :param address: Reference to a person's address.
        """
        return address.invalid

    def mark_address_invalid(self, person, address, invalid, **kwargs):
        """
        Mark the person's address as invalid.

        :param person: Reference to a person.
        :param address: Reference to a person's address.
        :param invalid: Boolean indicating "invalid" status for
           person's address.
        """
        address.invalid = invalid

    def resolve_person(self, pending, person, user, **kwargs):
        """
        Resolve a pending person.

        :param pending: Reference to a PendingCustomer instance.

        :param person: Reference to a Person instance.

        :param user: Reference to the User responsible.
        """
        custorder_handler = self.app.get_custorder_handler()
        custorder_handler.resolve_person(pending, person, user)

        pending.status_code = self.enum.PENDING_CUSTOMER_STATUS_RESOLVED

    def request_merge(self, user, removing_uuid, keeping_uuid, **kwargs):
        """
        Submit an officical merge request for two Person records.

        The caller must obviously specify which is to be kept and
        which removed, but really this is arbitrary, as the user
        performing the merge is free to swap them around.
        """
        model = self.model
        session = self.get_session(user)
        merge = model.MergePeopleRequest()
        merge.removing_uuid = removing_uuid
        merge.keeping_uuid = keeping_uuid
        merge.requested_by = user
        merge.requested = make_utc()
        session.add(merge)
        session.flush()
        self.notify_of_merge_request(merge)
        return merge

    def notify_of_merge_request(self, merge):
        """
        Send an email alert regarding a new merge request.
        """
        session = self.get_session(merge)
        model = self.model

        removing = session.get(model.Person, merge.removing_uuid)
        keeping = session.get(model.Person, merge.keeping_uuid)

        context = {
            'user_display': merge.requested_by.display_name,
            'removing_display': str(removing) if removing else "(not found)",
            'keeping_display': str(keeping) if keeping else "(not found)",
        }

        url = self.config.base_url()
        if url:
            context['merge_request_url'] = '{}/people/merge-requests/{}'.format(url, merge.uuid)
            if removing:
                context['removing_url'] = '{}/people/{}/profile'.format(url, removing.uuid)
            if keeping:
                context['keeping_url'] = '{}/people/{}/profile'.format(url, keeping.uuid)

        self.app.send_email('person_merge_request', context)

    def get_merge_preview_fields(self, **kwargs):
        """
        Returns a sequence of fields which will be used during a merge
        preview.
        """
        F = self.make_merge_field
        return [
            F('uuid'),
            F('first_name'),
            F('last_name'),
            F('display_name'),
            F('usernames', additive=True),
            F('employee_uuid', coalesce=True),
            F('member_uuids', additive=True),
        ]

    def get_merge_preview_data(self, person, **kwargs):
        """
        Must return a data dictionary for the given person, which can
        be presented to the user during a merge preview.
        """
        return {
            'uuid': person.uuid,
            'first_name': person.first_name,
            'last_name': person.last_name,
            'display_name': person.display_name,
            'usernames': [u.username for u in person.users],
            'employee_uuid': person.employee.uuid if person.employee else None,
            'member_uuids': [m.uuid for m in person.members],
        }

    def why_not_merge(self, removing, keeping, **kwargs):
        """
        Evaluate the given merge candidates and if there is a reason *not*
        to merge them, return that reason.

        :param removing: Person record which will be removed, should the
           merge happen.
        :param keeping: Person record which will be kept, should the
           merge happen.
        :returns: String indicating reason not to merge, or ``None``.
        """
        if removing.employee and keeping.employee:
            if removing.employee is not keeping.employee:
                return "Cannot merge 2 people who are distinct employees"

    def perform_merge(self, removing, keeping, **kwargs):

        # merge per usual
        super(PeopleHandler, self).perform_merge(removing, keeping, **kwargs)

        # if there were pending requests for this merge, declare them satisfied
        self.satisfy_merge_requests(removing, keeping, user=kwargs.get('user'))

    def merge_update_keeping_object(self, removing, keeping):

        # update per usual
        super(PeopleHandler, self).merge_update_keeping_object(removing, keeping)

        # move Member records to final Person
        for member in list(removing.members):
            removing.members.remove(member)
            keeping.members.append(member)

        # move Employee record to final Person
        if removing.employee:
            if keeping.employee and keeping.employee is not removing.employee:
                raise RuntimeError("Cannot merge 2 people who are distinct employees")
            if not keeping.employee:
                employee = removing.employee
                employee.person = keeping

        # move User records to final Person
        for user in list(removing.users):
            removing.users.remove(user)
            keeping.users.append(user)

    def satisfy_merge_requests(self, removing, keeping, user):
        """
        If there was a merge request(s) for this pair, mark it complete.
        """
        session = self.get_session(keeping)
        model = self.model
        merge_requests = session.query(model.MergePeopleRequest)\
                                .filter(sa.or_(
                                    sa.and_(
                                        model.MergePeopleRequest.removing_uuid == removing.uuid,
                                        model.MergePeopleRequest.keeping_uuid == keeping.uuid),
                                    sa.and_(
                                        model.MergePeopleRequest.removing_uuid == keeping.uuid,
                                        model.MergePeopleRequest.keeping_uuid == removing.uuid)))\
                                .all()
        for merge_request in merge_requests:
            # set the record straight re: removing vs. keeping
            merge_request.removing_uuid = removing.uuid
            merge_request.keeping_uuid = keeping.uuid
            merge_request.merged = make_utc()
            merge_request.merged_by = user

def get_people_handler(config, **kwargs):
    """
    Create and return the configured :class:`PeopleHandler` instance.

    .. warning::
       This function is deprecated; please use
       :meth:`~rattail.app.AppHandler.get_people_handler` instead.
    """
    warnings.warn("get_people_handler() function is deprecated, "
                  "please use app.get_people_handler() method instead",
                  DeprecationWarning)
    app = config.get_app()
    return app.get_people_handler(**kwargs)



# -*- coding: utf-8; -*-
"""fix equity_payments

Revision ID: 310080aaa346
Revises: 093db8c16e7d
Create Date: 2023-08-08 18:08:53.588048

"""

# revision identifiers, used by Alembic.
revision = '310080aaa346'
down_revision = '093db8c16e7d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # member_equity_payment_version
    op.alter_column('member_equity_payment_version', 'member_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=True)
    op.alter_column('member_equity_payment_version', 'amount',
               existing_type=sa.NUMERIC(precision=7, scale=2),
               nullable=True)
    op.alter_column('member_equity_payment_version', 'received',
               existing_type=postgresql.TIMESTAMP(),
               nullable=True)


def downgrade():

    # member_equity_payment_version
    op.alter_column('member_equity_payment_version', 'received',
               existing_type=postgresql.TIMESTAMP(),
               nullable=False)
    op.alter_column('member_equity_payment_version', 'amount',
               existing_type=sa.NUMERIC(precision=7, scale=2),
               nullable=False)
    op.alter_column('member_equity_payment_version', 'member_uuid',
               existing_type=sa.VARCHAR(length=32),
               nullable=False)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Autocomplete Handler for Customers
"""

from __future__ import unicode_literals, absolute_import

import re

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.autocomplete import Autocompleter
from rattail.autocomplete.base import PhoneMagicMixin
from rattail.db import model


class CustomerAutocompleter(Autocompleter):
    """
    Autocompleter for Customers (by name)
    """
    autocompleter_key = 'customers'
    model_class = model.Customer
    autocomplete_fieldname = 'name'


class CustomerPhoneAutocompleter(Autocompleter):
    """
    Autocompleter for Customers (by phone)

    .. note::
       As currently implemented, this view will only work with a
       PostgreSQL database.  It normalizes the user's search term and
       the database values to numeric digits only (i.e. removes
       special characters from each) in order to be able to perform
       smarter matching.  However normalizing the database value
       currently uses the PG SQL ``regexp_replace()`` function.
    """
    autocompleter_key = 'customers.phone'
    invalid_pattern = re.compile(r'\D')

    def prepare_autocomplete_term(self, term, **kwargs):
        return self.invalid_pattern.sub('', term)

    def make_autocomplete_query(self, session, term, **kwargs):
        model = self.model
        return session.query(model.CustomerPhoneNumber)\
                      .filter(sa.func.regexp_replace(model.CustomerPhoneNumber.number, r'\D', '', 'g').like('%{}%'.format(term)))\
                      .order_by(model.CustomerPhoneNumber.number)\
                      .options(orm.joinedload(model.CustomerPhoneNumber.customer))

    def autocomplete_display(self, phone):
        return "{} {}".format(phone.number, phone.customer)

    def autocomplete_value(self, phone):
        return phone.customer.uuid


class CustomerNewOrderAutocompleter(PhoneMagicMixin, Autocompleter):
    """
    Special "new order" autocompleter for customers.

    We set it apart with a different key (``'customers.neworder'``) so
    that you can override it independently of other customer
    autocompleters.

    But the default logic for this one is a bit special too, in that
    it will try to search for *either* phone number *or* customer
    name.  If the search term includes at least 4 digits then it is
    considered to be a phone number search; otherwise it will be
    considered a name search.
    """
    autocompleter_key = 'customers.neworder'
    model_class = model.Customer
    autocomplete_fieldname = 'name'
    phone_model_class = model.CustomerPhoneNumber



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Cleanup Handler

See also :doc:`rattail-manual:base/handlers/other/cleanup`.
"""

from __future__ import unicode_literals, absolute_import

import logging

from rattail.app import GenericHandler
from rattail.util import load_entry_points


log = logging.getLogger(__name__)


class CleanupHandler(GenericHandler):
    """
    Base class and default implementation for the "cleanup" handler,
    responsible for removing old unwanted files etc.
    """

    def cleanup_everything(self, session, dry_run=False, progress=None, **kwargs):
        """
        Invoke cleanup logic for all enabled cleaners.
        """
        cleaners = self.get_all_cleaners()
        for key in sorted(cleaners):
            cleaner = cleaners[key]
            log.debug("running cleanup for: %s", key)
            cleaner.cleanup(session, dry_run=dry_run, progress=progress)

    def get_all_cleaners(self, **kwargs):
        """
        Return a dictionary containing all registered cleaner objects.
        """
        cleaners = load_entry_points('rattail.cleaners')
        for key in list(cleaners):
            cleaner = cleaners[key](self.config)
            cleaner.key = key
            cleaners[key] = cleaner
        return cleaners

    def get_cleaner(self, key):
        """
        Retrieve a specific cleaner object.
        """
        cleaners = self.get_all_cleaners()
        return cleaners.get(key)


class Cleaner(object):
    """
    Base class for cleaners.
    """

    def __init__(self, config, **kwargs):
        self.config = config
        self.app = config.get_app()
        self.model = self.app.model

    def cleanup(self, session, dry_run=False, progress=None, **kwargs):
        """
        Perform actual cleanup steps as needed.
        """



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Watchers
"""

import datetime

import sqlalchemy as sa

from rattail.time import make_utc, localtime


class DataSyncWatcher(object):
    """
    Base class for all DataSync watchers.
    """
    prunes_changes = False
    retry_attempts = 1
    retry_delay = 1 # seconds

    def __init__(self, config, key, dbkey=None, **kwargs):
        """
        Constructor.

        Note that while arbitrary kwargs are allowed, the default constructor
        method does *not* process these in any way.  So if you need to accept
        some kwargs then you must also define the processing thereof.

        The primary reason for this is because these kwargs, if present, will
        *always* be simple strings due to how they are read from config.
        Therefore the burden is on each custom watcher, to interpret them as
        whatever data type is necessary.
        """
        self.config = config
        self.key = key
        self.dbkey = dbkey
        self.delay = 1 # seconds
        self.model = self.config.get_model()
        self.app = self.config.get_app()

    def setup(self):
        """
        This method is called when the watcher thread is first started.
        """

    def localize_lastrun(self, session, lastrun):
        """
        Calculates a timestamp using `lastrun` as the starting point, but also
        taking into account a possible (?) time drift between the local and
        "other" server.
        """
        # get current time according to "other" server, and convert to UTC
        before = make_utc()
        other_now = session.execute(sa.text('select current_timestamp'))\
                           .fetchone()[0]
        after = make_utc()
        other_now = make_utc(localtime(self.config, other_now))

        # get current time according to local server (i.e. Rattail), in UTC
        local_now = after

        # drift is essentially the difference between the two timestamps.  it
        # is meant to be a positive value (or zero) since it will be
        # "subtracted" from the `lastrun` time in order to obtain the timestamp
        # we should use for "other" queries.  note that we also add one second
        # to the drift, just to be on the safe side.
        if other_now < local_now:
            drift = local_now - other_now
        else:
            drift = datetime.timedelta(seconds=0)
        drift += datetime.timedelta(seconds=1)

        # convert lastrun time to local timezone, for "other" queries
        lastrun = localtime(self.config, lastrun, tzinfo=False)

        # and finally, apply the drift
        return lastrun - drift

    def get_changes(self, lastrun):
        """
        This must be implemented by the subclass.  It should check the source
        database for pending changes, and return a list of
        :class:`rattail.db.model.DataSyncChange` instances representing the
        source changes.
        """
        return []

    def prune_changes(self, keys):
        """
        Prune change records from the source database, if relevant.
        """

    def process_changes(self, session, changes):
        """
        Process (consume) a batch of changes.
        """


class NullWatcher(DataSyncWatcher):
    """
    Null watcher, will never actually check for or report any changes.
    """


class ErrorTestWatcher(DataSyncWatcher):
    """
    Watcher which always raises an error when attempting to get changes.
    Useful for testing error handling etc.
    """

    def get_changes(self, lastrun):
        raise RuntimeError("Fake exception, to test error handling")



# -*- coding: utf-8; -*-
"""add pack_size for delproduct batch

Revision ID: 9072c76a3afc
Revises: b47eeb5fcf79
Create Date: 2021-07-15 12:40:18.374163

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '9072c76a3afc'
down_revision = 'b47eeb5fcf79'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_delproduct_row
    op.add_column('batch_delproduct_row', sa.Column('pack_size', sa.Numeric(precision=9, scale=4), nullable=True))


def downgrade():

    # batch_delproduct_row
    op.drop_column('batch_delproduct_row', 'pack_size')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CSV -> Rattail data import
"""

import os
import csv
import datetime
import decimal
from collections import OrderedDict

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy_utils.functions import get_primary_keys

from rattail import importing, csvutil
from rattail.importing.handlers import FromFileHandler
from rattail.importing.files import FromFile
from rattail.db.util import make_topo_sortkey
from rattail.config import parse_bool


class FromCSVToSQLAlchemyMixin(object):

    host_key = 'csv'
    generic_host_title = "CSV"

    # subclass must define this
    ToParent = None

    def get_importers(self):
        """
        Here we build the set of available importers on the fly.  This avoids
        having to define things over and over since really we're just going to
        piggy-back on the existing logic, for storing new data.
        """
        importers = {}
        model = self.get_model()

        # mostly try to make an importer for every data model
        for name in dir(model):
            obj = getattr(model, name, None)
            if isinstance(obj, type) and issubclass(obj, model.Base) and obj is not model.Base:
                importers[name] = self.make_importer_factory(name, obj)

        # sort importers according to topography
        topo_sortkey = make_topo_sortkey(model)
        importers = OrderedDict([
            (name, importers[name])
            for name in sorted(importers, key=topo_sortkey)
        ])

        return importers

    def make_importer_factory(self, name, cls):
        mapper = orm.class_mapper(cls)
        fields = list(mapper.columns.keys())
        pkeys = get_primary_keys(cls)
        name = '{}Importer'.format(name)
        return type(name, (FromCSV, self.ToParent), {
            'model_class': cls,
            'supported_fields': fields,
            'simple_fields': fields,
            'key': list(pkeys),
            'coercers': self.make_coercers(cls, fields),
        })

    def make_coercers(self, cls, fields):
        coercers = {}
        for field in fields:
            attr = getattr(cls, field)
            assert len(attr.prop.columns) == 1
            column = attr.prop.columns[0]

            # String
            if isinstance(attr.type, sa.String):
                if column.nullable:
                    coercers[field] = self.coerce_string_nullable
                else:
                    coercers[field] = self.coerce_string

            # Boolean
            elif isinstance(attr.type, sa.Boolean):
                if column.nullable:
                    coercers[field] = self.coerce_boolean_nullable
                else:
                    coercers[field] = self.coerce_boolean

            # Integer
            elif isinstance(attr.type, sa.Integer):
                if column.nullable:
                    coercers[field] = self.coerce_integer_nullable
                else:
                    coercers[field] = self.coerce_integer

            # Float
            elif isinstance(attr.type, sa.Float):
                if column.nullable:
                    coercers[field] = self.coerce_float_nullable
                else:
                    coercers[field] = self.coerce_float

            # Decimal
            elif isinstance(attr.type, sa.Numeric):
                if column.nullable:
                    coercers[field] = self.coerce_decimal_nullable
                else:
                    coercers[field] = self.coerce_decimal

            # DateTime
            elif (isinstance(attr.type, sa.DateTime)
                  or (hasattr(attr.type, 'impl')
                      and isinstance(attr.type.impl, sa.DateTime))):
                if column.nullable:
                    coercers[field] = self.coerce_datetime_nullable
                else:
                    coercers[field] = self.coerce_datetime

            else: # unknown type; do not coerce
                coercers[field] = lambda value: value

        return coercers

    def coerce_boolean(self, value):
        return parse_bool(value)

    def coerce_boolean_nullable(self, value):
        if value == '':
            return None
        return self.coerce_boolean(value)

    def coerce_integer(self, value):
        if value == '':
            return None
        return int(value)

    def coerce_integer_nullable(self, value):
        if value == '':
            return None
        return self.coerce_integer(value)

    def coerce_float(self, value):
        if value == '':
            return None
        return float(value)

    def coerce_float_nullable(self, value):
        if value == '':
            return None
        return self.coerce_float(value)

    def coerce_decimal(self, value):
        if value == '':
            return None
        return decimal.Decimal(value)

    def coerce_decimal_nullable(self, value):
        if value == '':
            return None
        return self.coerce_decimal(value)

    def coerce_datetime(self, value):
        if value == '':
            return

        try:
            return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        except ValueError:
            return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f')

    def coerce_datetime_nullable(self, value):
        if value == '':
            return None
        return self.coerce_datetime(value)

    def coerce_string(self, value):
        return value

    def coerce_string_nullable(self, value):
        if value == '':
            return None
        return self.coerce_string(value)


class FromCSVToRattail(FromCSVToSQLAlchemyMixin, FromFileHandler, importing.ToRattailHandler):
    """
    Handler for CSV -> Rattail data import
    """
    host_title = "CSV"
    ToParent = importing.model.ToRattail

    @property
    def local_title(self):
        return self.config.node_title()

    def get_model(self):
        if self.config:
            return self.config.get_model()

        from rattail.db import model
        return model


class FromCSV(FromFile):
    """
    Base class for importers coming from CSV
    """
    csv_encoding = 'utf_8'

    def get_input_file_name(self):
        return '{}.csv'.format(self.model_name)

    def open_input_file(self):
        self.input_file = open(self.input_file_path, 'rt', encoding=self.csv_encoding)
        self.input_reader = csv.DictReader(self.input_file)

    def close_input_file(self):
        self.input_file.close()

    def get_host_objects(self):
        return list(self.input_reader)

    def normalize_host_object(self, data):
        return self.coerce_local(data)

    def coerce_local(self, data):
        coerced = {}
        for field in self.fields:
            value = data[field]
            coerced[field] = self.coercers[field](value)
        return coerced



# -*- coding: utf-8; -*-
"""add MemberEquityPayment

Revision ID: e00841e2deb2
Revises: a4ac7985cfa6
Create Date: 2023-06-18 11:59:41.256360

"""

# revision identifiers, used by Alembic.
revision = 'e00841e2deb2'
down_revision = 'a4ac7985cfa6'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # member_equity_payment
    op.create_table('member_equity_payment',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('member_uuid', sa.String(length=32), nullable=False),
                    sa.Column('amount', sa.Numeric(precision=7, scale=2), nullable=False),
                    sa.Column('received', sa.DateTime(), nullable=False),
                    sa.Column('description', sa.String(length=100), nullable=True),
                    sa.Column('transaction_identifier', sa.String(length=100), nullable=True),
                    sa.ForeignKeyConstraint(['member_uuid'], ['member.uuid'], name='member_equity_payment_fk_member'),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('member_equity_payment_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('member_uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('amount', sa.Numeric(precision=7, scale=2), autoincrement=False, nullable=False),
                    sa.Column('received', sa.DateTime(), autoincrement=False, nullable=False),
                    sa.Column('description', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_identifier', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_member_equity_payment_version_end_transaction_id'), 'member_equity_payment_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_member_equity_payment_version_operation_type'), 'member_equity_payment_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_member_equity_payment_version_transaction_id'), 'member_equity_payment_version', ['transaction_id'], unique=False)


def downgrade():

    # member_equity_payment
    op.drop_index(op.f('ix_member_equity_payment_version_transaction_id'), table_name='member_equity_payment_version')
    op.drop_index(op.f('ix_member_equity_payment_version_operation_type'), table_name='member_equity_payment_version')
    op.drop_index(op.f('ix_member_equity_payment_version_end_transaction_id'), table_name='member_equity_payment_version')
    op.drop_table('member_equity_payment_version')
    op.drop_table('member_equity_payment')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Handler
"""

from __future__ import unicode_literals, absolute_import

import datetime
import logging
import subprocess

from six.moves.xmlrpc_client import ProtocolError

from rattail.app import GenericHandler
from rattail.datasync.config import load_profiles


log = logging.getLogger(__name__)


class DatasyncHandler(GenericHandler):

    def should_use_profile_settings(self):
        """
        This declares whether datasync profiles should be read from
        config file only (the default) or from settings in the DB.

        :returns: True if settings in the DB should be used; False if not.
        """
        return self.config.getbool('rattail.datasync', 'use_profile_settings',
                                   default=False)

    def get_supervisor_process_name(self, require=False, **kwargs):
        getter = self.config.require if require else self.config.get
        return getter('rattail.datasync', 'supervisor_process_name')

    def get_supervisor_process_info(self, name=None, **kwargs):
        if not name:
            name = self.get_supervisor_process_name(require=True)

        proxy = self.app.make_supervisorctl_proxy()

        try:
            return proxy.supervisor.getProcessInfo(name)
        except ProtocolError as error:
            raise self.app.safe_supervisor_protocol_error(error)

    def restart_supervisor_process(self, name=None, **kwargs):
        if not name:
            name = self.get_supervisor_process_name()

        try:
            proxy = self.app.make_supervisorctl_proxy()
        except:
            log.warning("failed to make supervisorctl proxy", exc_info=True)

        else:
            # we have our proxy, so use that, then return
            try:
                info = proxy.supervisor.getProcessInfo(name)
                if info['state'] != 0:
                    proxy.supervisor.stopProcess(name)
                proxy.supervisor.startProcess(name)
            except ProtocolError as error:
                raise self.app.safe_supervisor_protocol_error(error)
            return

        # no proxy, but we can still try command line
        # TODO: should rename this setting at some point?
        cmd = self.config.get('tailbone', 'datasync.restart')
        if cmd:
            cmd = self.config.parse_list(cmd)
        elif name:
            cmd = ['supervisorctl', 'restart', name]

        log.debug("attempting datasync restart with command: %s", cmd)

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as error:
            log.warning("failed to restart datasync; output was:")
            log.warning(error.output)
            raise

    def get_configured_profiles(self, **kwargs):
        return load_profiles(self.config, **kwargs)

    def get_watcher_lastrun(self, watcher_key, local=False, tzinfo=True,
                            session=None):
        """
        Retrieve the "last run" time for the datasync watcher thread
        identifed by ``watcher_key``.
        """
        # the 'last run' value is maintained as UTC
        lastrun_setting = self.get_watcher_lastrun_setting(watcher_key)
        timefmt = self.get_lastrun_timefmt()

        close = False
        if not session:
            session = self.app.make_session()
            close = True
        lastrun = self.app.get_setting(session, lastrun_setting)
        if close:
            session.close()

        if lastrun:
            lastrun = datetime.datetime.strptime(lastrun, timefmt)
            if local:
                return self.app.localtime(lastrun, from_utc=True, tzinfo=tzinfo)
            else:
                return self.app.make_utc(lastrun, tzinfo=tzinfo)

    def get_lastrun_timefmt(self):
        """
        Retrieve the "last run" time format.
        """
        return '%Y-%m-%d %H:%M:%S'

    def get_watcher_lastrun_setting(self, watcher_key):
        """
        Retrieve the "last run" setting name for the given watcher.
        """
        return 'rattail.datasync.{}.watcher.lastrun'.format(watcher_key)



# -*- coding: utf-8; -*-
"""add item discounts

Revision ID: 811f436d593c
Revises: c3b721c34f4b
Create Date: 2022-02-19 15:27:24.452125

"""

from __future__ import unicode_literals, absolute_import

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '811f436d593c'
down_revision = 'c3b721c34f4b'
branch_labels = None
depends_on = None


def upgrade():

    # transaction_item_discount
    op.create_table('transaction_item_discount',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('item_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=50), nullable=False),
                    sa.Column('amount', sa.Numeric(precision=9, scale=2), nullable=False),
                    sa.ForeignKeyConstraint(['item_uuid'], ['transaction_item.uuid'], name='transaction_item_discount_fk_item'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # transaction_item_discount
    op.drop_table('transaction_item_discount')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Autocomplete Handler for Members
"""

import sqlalchemy as sa

from rattail.autocomplete import Autocompleter
from rattail.db.model import Member


class MemberAutocompleter(Autocompleter):
    """
    Autocompleter for Members (by name)
    """
    autocompleter_key = 'members'
    model_class = Member
    autocomplete_fieldname = 'name'

    def make_base_query(self, session):
        model = self.model
        return session.query(model.Member)\
                      .outerjoin(model.Customer)\
                      .outerjoin(model.Person,
                                 model.Person.uuid == model.Member.person_uuid)

    def filter_autocomplete_query(self, session, query, term):
        model = self.model

        # TODO: should also filter by Customer.name ?
        # or maybe just need a separate autocompleter for that
        column = model.Person.display_name
        criteria = [column.ilike(f'%{word}%')
                    for word in term.split()]
        return query.filter(sa.and_(*criteria))

    def autocomplete_display(self, member):
        return member.person.display_name

    def sort_autocomplete_query(self, session, query):
        model = self.model
        return query.order_by(model.Person.display_name)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail data import
"""

import logging
from collections import OrderedDict

import sqlalchemy as sa

from rattail.db import Session
from rattail.importing import model
from rattail.importing.handlers import FromSQLAlchemyHandler, ToSQLAlchemyHandler
from rattail.importing.sqlalchemy import FromSQLAlchemySameToSame


log = logging.getLogger(__name__)


class FromRattailHandler(FromSQLAlchemyHandler):
    """
    Base class for import handlers which target a Rattail database on the local side.
    """
    host_key = 'rattail'
    host_title = "Rattail"
    generic_host_title = "Rattail"

    @property
    def host_title(self):
        return self.config.app_title(default="Rattail")

    def make_host_session(self):
        return Session()


class ToRattailHandler(ToSQLAlchemyHandler):
    """
    Base class for import handlers which target a Rattail database on the local side.
    """
    generic_local_title = "Rattail"
    local_key = 'rattail'

    @property
    def local_title(self):
        return self.config.app_title(default="Rattail")

    def make_session(self):
        kwargs = {}
        if hasattr(self, 'runas_user'):
            kwargs['continuum_user'] = self.runas_user
        return Session(**kwargs)

    def begin_local_transaction(self):
        self.session = self.make_session()

        # load "runas user" into current session
        if hasattr(self, 'runas_user') and self.runas_user:
            dbmodel = self.config.get_model()
            runas_user = self.session.get(dbmodel.User, self.runas_user.uuid)
            if not runas_user:
                log.info("runas_user does not exist in target session: %s",
                         self.runas_user.username)
            # this may be None if user does not exist in target session
            self.runas_user = runas_user

        # declare "runas user" is data versioning author
        if hasattr(self, 'runas_username') and self.runas_username:
            self.session.set_continuum_user(self.runas_username)


class FromRattailToRattailBase(object):
    """
    Common base class for Rattail -> Rattail data import/export handlers.
    """

    def get_importers(self):
        importers = OrderedDict()
        importers['Person'] = PersonImporter
        importers['GlobalPerson'] = GlobalPersonImporter
        importers['PersonEmailAddress'] = PersonEmailAddressImporter
        importers['PersonPhoneNumber'] = PersonPhoneNumberImporter
        importers['PersonMailingAddress'] = PersonMailingAddressImporter
        importers['MergePeopleRequest'] = MergePeopleRequestImporter
        importers['Role'] = RoleImporter
        importers['GlobalRole'] = GlobalRoleImporter
        importers['User'] = UserImporter
        importers['AdminUser'] = AdminUserImporter
        importers['GlobalUser'] = GlobalUserImporter
        importers['Message'] = MessageImporter
        importers['MessageRecipient'] = MessageRecipientImporter
        importers['Store'] = StoreImporter
        importers['StorePhoneNumber'] = StorePhoneNumberImporter
        importers['Employee'] = EmployeeImporter
        importers['EmployeeStore'] = EmployeeStoreImporter
        importers['EmployeeEmailAddress'] = EmployeeEmailAddressImporter
        importers['EmployeePhoneNumber'] = EmployeePhoneNumberImporter
        importers['ScheduledShift'] = ScheduledShiftImporter
        importers['WorkedShift'] = WorkedShiftImporter
        importers['Customer'] = CustomerImporter
        importers['CustomerGroup'] = CustomerGroupImporter
        importers['CustomerGroupAssignment'] = CustomerGroupAssignmentImporter
        importers['CustomerShopper'] = CustomerShopperImporter
        importers['CustomerShopperHistory'] = CustomerShopperHistoryImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['CustomerEmailAddress'] = CustomerEmailAddressImporter
        importers['CustomerPhoneNumber'] = CustomerPhoneNumberImporter
        importers['Member'] = MemberImporter
        importers['MemberEmailAddress'] = MemberEmailAddressImporter
        importers['MemberPhoneNumber'] = MemberPhoneNumberImporter
        importers['MemberEquityPayment'] = MemberEquityPaymentImporter
        importers['Vendor'] = VendorImporter
        importers['VendorEmailAddress'] = VendorEmailAddressImporter
        importers['VendorPhoneNumber'] = VendorPhoneNumberImporter
        importers['VendorContact'] = VendorContactImporter
        importers['VendorSampleFile'] = VendorSampleFileImporter
        importers['Department'] = DepartmentImporter
        importers['EmployeeDepartment'] = EmployeeDepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Category'] = CategoryImporter
        importers['Family'] = FamilyImporter
        importers['ReportCode'] = ReportCodeImporter
        importers['DepositLink'] = DepositLinkImporter
        importers['Tax'] = TaxImporter
        importers['InventoryAdjustmentReason'] = InventoryAdjustmentReasonImporter
        importers['Brand'] = BrandImporter
        importers['Product'] = ProductImporter
        importers['ProductCode'] = ProductCodeImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductPrice'] = ProductPriceImporter
        importers['ProductPriceAssociation'] = ProductPriceAssociationImporter
        importers['ProductStoreInfo'] = ProductStoreInfoImporter
        importers['ProductVolatile'] = ProductVolatileImporter
        importers['ProductImage'] = ProductImageImporter
        importers['LabelProfile'] = LabelProfileImporter
        return importers

    def get_default_keys(self):
        keys = self.get_importer_keys()

        avoid_by_default = [
            'Role',
            'GlobalRole',
            'GlobalPerson',
            'AdminUser',
            'GlobalUser',
            'ProductImage',
            'ProductPriceAssociation',
        ]

        for key in avoid_by_default:
            if key in keys:
                keys.remove(key)

        return keys


class FromRattailToRattailImport(FromRattailToRattailBase, FromRattailHandler, ToRattailHandler):
    """
    Handler for Rattail (other) -> Rattail (local) data import.

    .. attribute:: direction

       Value is ``'import'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    dbkey = 'other'

    @property
    def host_title(self):
        return "{} ({})".format(self.config.app_title(default="Rattail"), self.dbkey)

    @property
    def local_title(self):
        return self.config.node_title(default="Rattail (local)")

    def make_host_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])


class FromRattailToRattailExport(FromRattailToRattailBase, FromRattailHandler, ToRattailHandler):
    """
    Handler for Rattail (local) -> Rattail (other) data export.

    .. attribute:: direction

       Value is ``'export'`` - see also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.
    """
    direction = 'export'
    dbkey = 'other'

    @property
    def host_title(self):
        return self.config.node_title()

    @property
    def local_title(self):
        return "{} ({})".format(self.config.app_title(default="Rattail"), self.dbkey)

    def make_session(self):
        return Session(bind=self.config.rattail_engines[self.dbkey])


class FromRattail(FromSQLAlchemySameToSame):
    """
    Base class for Rattail -> Rattail data importers.
    """


class PersonImporter(FromRattail, model.PersonImporter):
    pass


class GlobalPersonImporter(FromRattail, model.GlobalPersonImporter):
    """
    This is a customized version of the :class:`PersonImporter`, which simply
    avoids "local only" person accounts.
    """

    def query(self):
        query = super(GlobalPersonImporter, self).query()

        # never include "local only" people
        query = query.filter(sa.or_(
            self.host_model_class.local_only == False,
            self.host_model_class.local_only == None))

        return query

    def normalize_host_object(self, person):

        # must check this here for sake of datasync
        if person.local_only:
            return

        data = super(GlobalPersonImporter, self).normalize_host_object(person)
        return data


class PersonEmailAddressImporter(FromRattail, model.PersonEmailAddressImporter):
    pass

class PersonPhoneNumberImporter(FromRattail, model.PersonPhoneNumberImporter):
    pass

class PersonMailingAddressImporter(FromRattail, model.PersonMailingAddressImporter):
    pass

class MergePeopleRequestImporter(FromRattail, model.MergePeopleRequestImporter):
    pass

class RoleImporter(FromRattail, model.RoleImporter):
    pass


class GlobalRoleImporter(RoleImporter):
    """
    Role importer which only will handle roles which have the
    :attr:`~rattail.db.model.users.Role.sync_me` flag set.  (So it
    syncs those roles but avoids others.)
    """

    @property
    def supported_fields(self):
        fields = list(super(GlobalRoleImporter, self).supported_fields)
        fields.extend([
            'permissions',
            'users',
        ])
        return fields

    # nb. we must override both cache_query() and query() b/c they use
    # different sessions!

    def cache_query(self):
        """
        Return the query to be used when caching "local" data.
        """
        query = super(GlobalRoleImporter, self).cache_query()
        model = self.model

        # only want roles which are *meant* to be synced
        query = query.filter(model.Role.sync_me == True)

        return query

    def query(self):
        query = super(GlobalRoleImporter, self).query()
        model = self.model

        # only want roles which are *meant* to be synced
        query = query.filter(model.Role.sync_me == True)

        return query

    # nb. we do not need to override normalize_host_object() b/c it
    # just calls normalize_local_object() by default

    def normalize_local_object(self, role):

        # only want roles which are *meant* to be synced
        if not role.sync_me:
            return

        data = super(GlobalRoleImporter, self).normalize_local_object(role)
        if data:

            # users
            if 'users' in self.fields:
                data['users'] = sorted([user.uuid for user in role.users])

            # permissions
            if 'permissions' in self.fields:
                auth = self.app.get_auth_handler()
                perms = auth.cache_permissions(self.session, role,
                                               include_guest=False)
                data['permissions'] = sorted(perms)

            return data

    def update_object(self, role, host_data, local_data=None, **kwargs):
        role = super(GlobalRoleImporter, self).update_object(role, host_data,
                                                             local_data=local_data,
                                                             **kwargs)
        model = self.model

        # users
        # nb. we only update users if this role has flag set
        if 'users' in self.fields and role.sync_users:

            new_users = host_data['users']
            old_users = local_data['users'] if local_data else []
            changed = False

            # add some users
            for new_user in new_users:
                if new_user not in old_users:
                    user = self.session.get(model.User, new_user)
                    if user:
                        user.roles.append(role)
                        changed = True

            # remove some users
            for old_user in old_users:
                if old_user not in new_users:
                    user = self.session.get(model.User, old_user)
                    if user:
                        user.roles.remove(role)
                        changed = True

            if changed:
                self.session.flush()
                self.session.refresh(role)
                # also record a change to the role, for datasync.
                # this is done "just in case" the role is to be
                # synced to all nodes
                if self.session.rattail_record_changes:
                    self.session.add(model.Change(class_name='Role',
                                                  instance_uuid=role.uuid,
                                                  deleted=False))

        # permissions
        if 'permissions' in self.fields:
            auth = self.app.get_auth_handler()
            new_perms = host_data['permissions']
            old_perms = local_data['permissions'] if local_data else []

            # grant permissions
            for new_perm in new_perms:
                if new_perm not in old_perms:
                    auth.grant_permission(role, new_perm)

            # revoke permissions
            for old_perm in old_perms:
                if old_perm not in new_perms:
                    auth.revoke_permission(role, old_perm)

        return role


class UserImporter(FromRattail, model.UserImporter):
    pass


class GlobalUserImporter(FromRattail, model.GlobalUserImporter):
    """
    This is a customized version of the :class:`UserImporter`, which simply
    avoids "local only" user accounts.
    """

    def query(self):
        query = super(GlobalUserImporter, self).query()

        # never include "local only" users
        query = query.filter(sa.or_(
            self.host_model_class.local_only == False,
            self.host_model_class.local_only == None))

        return query

    def normalize_host_object(self, user):

        # must check this here for sake of datasync
        if user.local_only:
            return

        data = super(GlobalUserImporter, self).normalize_host_object(user)
        return data


class AdminUserImporter(FromRattail, model.AdminUserImporter):

    @property
    def supported_fields(self):
        return super(AdminUserImporter, self).supported_fields + [
            'admin',
        ]

    def normalize_host_object(self, user):
        data = super(AdminUserImporter, self).normalize_local_object(user) # sic
        if 'admin' in self.fields: # TODO: do we really need this, after the above?
            data['admin'] = self.admin_uuid in [r.role_uuid for r in user._roles]
        return data


class MessageImporter(FromRattail, model.MessageImporter):
    pass

class MessageRecipientImporter(FromRattail, model.MessageRecipientImporter):
    pass

class StoreImporter(FromRattail, model.StoreImporter):
    pass

class StorePhoneNumberImporter(FromRattail, model.StorePhoneNumberImporter):
    pass

class EmployeeImporter(FromRattail, model.EmployeeImporter):
    pass

class EmployeeStoreImporter(FromRattail, model.EmployeeStoreImporter):
    pass

class EmployeeDepartmentImporter(FromRattail, model.EmployeeDepartmentImporter):
    pass

class EmployeeEmailAddressImporter(FromRattail, model.EmployeeEmailAddressImporter):
    pass

class EmployeePhoneNumberImporter(FromRattail, model.EmployeePhoneNumberImporter):
    pass

class ScheduledShiftImporter(FromRattail, model.ScheduledShiftImporter):
    pass

class WorkedShiftImporter(FromRattail, model.WorkedShiftImporter):
    pass

class CustomerImporter(FromRattail, model.CustomerImporter):
    pass

class CustomerGroupImporter(FromRattail, model.CustomerGroupImporter):
    pass

class CustomerGroupAssignmentImporter(FromRattail, model.CustomerGroupAssignmentImporter):
    pass

class CustomerShopperImporter(FromRattail, model.CustomerShopperImporter):
    pass

class CustomerShopperHistoryImporter(FromRattail, model.CustomerShopperHistoryImporter):
    pass

class CustomerPersonImporter(FromRattail, model.CustomerPersonImporter):
    pass

class CustomerEmailAddressImporter(FromRattail, model.CustomerEmailAddressImporter):
    pass

class CustomerPhoneNumberImporter(FromRattail, model.CustomerPhoneNumberImporter):
    pass

class MemberImporter(FromRattail, model.MemberImporter):
    pass

class MemberEmailAddressImporter(FromRattail, model.MemberEmailAddressImporter):
    pass

class MemberPhoneNumberImporter(FromRattail, model.MemberPhoneNumberImporter):
    pass

class MemberEquityPaymentImporter(FromRattail, model.MemberEquityPaymentImporter):
    pass

class VendorImporter(FromRattail, model.VendorImporter):
    pass

class VendorEmailAddressImporter(FromRattail, model.VendorEmailAddressImporter):
    pass

class VendorPhoneNumberImporter(FromRattail, model.VendorPhoneNumberImporter):
    pass

class VendorContactImporter(FromRattail, model.VendorContactImporter):
    pass

class VendorSampleFileImporter(FromRattail, model.VendorSampleFileImporter):
    pass

class DepartmentImporter(FromRattail, model.DepartmentImporter):
    pass

class SubdepartmentImporter(FromRattail, model.SubdepartmentImporter):
    pass

class CategoryImporter(FromRattail, model.CategoryImporter):
    pass

class FamilyImporter(FromRattail, model.FamilyImporter):
    pass

class ReportCodeImporter(FromRattail, model.ReportCodeImporter):
    pass

class DepositLinkImporter(FromRattail, model.DepositLinkImporter):
    pass

class TaxImporter(FromRattail, model.TaxImporter):
    pass

class InventoryAdjustmentReasonImporter(FromRattail, model.InventoryAdjustmentReasonImporter):
    pass

class BrandImporter(FromRattail, model.BrandImporter):
    pass


class ProductWithPriceImporter(FromRattail, model.ProductImporter):
    """
    This can perhaps be thought of as the "complete" Product record
    importer.  The "normal" Product importer will typically avoid the
    "price uuid" reference fields, b/c of that foreign key chaos.

    Note that this importer is not (yet?) used directly, but is
    primarily useful as a base class.
    """
    # these require special handling due to the 2-way table dependency
    price_reference_fields = [
        'regular_price_uuid',
        'tpr_price_uuid',
        'sale_price_uuid',
        'current_price_uuid',
        'suggested_price_uuid',
    ]

    def query(self):
        query = super(ProductWithPriceImporter, self).query()

        # make sure potential unit items (i.e. rows with NULL unit_uuid) come
        # first, so they will be created before pack items reference them
        # cf. https://www.postgresql.org/docs/current/static/queries-order.html
        # cf. https://stackoverflow.com/a/7622046
        query = query.order_by(self.host_model_class.unit_uuid.desc())

        return query


class ProductPriceAssociationImporter(ProductWithPriceImporter):
    """
    Note that this importer is *only* for sake of handling the "price
    uuid" fields.
    """

    @property
    def simple_fields(self):
        return ['uuid'] + self.price_reference_fields


class ProductImporter(ProductWithPriceImporter):
    """
    Note that this is the "normal" Product record importer, but it
    inherits from the "complete" importer.  This one avoids the "price
    uuid" fields to avoid that foreign key chaos.
    """

    @property
    def simple_fields(self):
        fields = super(ProductImporter, self).simple_fields
        # NOTE: it seems we can't consider these "simple" due to the
        # self-referencing foreign key situation.  an importer can still
        # "support" these fields, but they're excluded from the simple set for
        # sake of rattail <-> rattail
        for field in self.price_reference_fields:
            fields.remove(field)
        return fields


class ProductCodeImporter(FromRattail, model.ProductCodeImporter):
    pass

class ProductCostImporter(FromRattail, model.ProductCostImporter):
    pass

class ProductPriceImporter(FromRattail, model.ProductPriceImporter):

    @property
    def supported_fields(self):
        # nb. parent class FromRattail only supports simple_fields, so
        # we explicitly copy logic from model importer class here.
        return self.simple_fields + self.product_reference_fields


class ProductStoreInfoImporter(FromRattail, model.ProductStoreInfoImporter):
    pass

class ProductVolatileImporter(FromRattail, model.ProductVolatileImporter):
    pass


class ProductImageImporter(FromRattail, model.ProductImageImporter):
    """
    Importer for product images.  Note that this uses the "batch" approach
    because fetching all data up front is not performant when the host/local
    systems are on different machines etc.
    """

    def query(self):
        query = self.host_session.query(self.model_class)\
                                 .order_by(self.model_class.uuid)
        return query[self.host_index:self.host_index + self.batch_size]


class LabelProfileImporter(FromRattail, model.LabelProfileImporter):

    def query(self):
        query = super(LabelProfileImporter, self).query()

        if not self.config.getbool('rattail', 'labels.sync_all_profiles', default=False):
            # only fetch labels from host which are marked as "sync me"
            query = query .filter(self.model_class.sync_me == True)

        return query.order_by(self.model_class.ordinal)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Luigi Logging
"""

from __future__ import unicode_literals, absolute_import

import sys
import logging


class WarnSummaryAlways(logging.Filter):
    """
    Custom logging filter, to elevate the Luigi "execution summary"
    message to WARNING level, so that it shows up even when we ignore
    INFO logging.
    """

    def filter(self, record):
        if record.name == 'luigi-interface' and record.levelno == logging.INFO:
            if "===== Luigi Execution Summary =====" in record.msg:

                # looks like a summary, raise to WARNING
                record.levelno = logging.WARNING
                record.levelname = 'WARNING'

                # nb. also emit a blank line, to help set summary
                # apart from any previous output
                sys.stderr.write("\n")

        return True


class WarnSummaryIfProblems(logging.Filter):
    """
    Custom logging filter, to elevate the Luigi "execution summary" message to
    WARNING level, if any problems are detected for the run.  Note that this
    simply checks for the ``:)`` message to know if there were problem.
    """

    good_messages = [
        "This progress looks :) because there were no failed tasks or missing external dependencies",
        "This progress looks :) because there were no failed tasks or missing dependencies",
    ]


    def filter(self, record):
        if record.name == 'luigi-interface' and record.levelno == logging.INFO:
            if "===== Luigi Execution Summary =====" in record.msg:
                if not any([msg in record.msg for msg in self.good_messages]):
                    record.levelno = logging.WARNING
                    record.levelname = 'WARNING'
        return True



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
"Customer Mailing" report
"""

from sqlalchemy import orm

from rattail.reporting import ExcelReport


class CustomerMailing(ExcelReport):
    """
    Generates a customer mailing address list.
    """
    type_key = 'customer_mailing'
    name = "Customer Mailing"

    output_fields = [
        'first_name',
        'last_name',
        'street',
        'street2',
        'city',
        'state',
        'zipcode',
        'address_invalid',
    ]

    def make_data(self, session, params, progress=None, **kwargs):
        model = self.model

        # looking for all customers with account holder
        customers = session.query(model.Customer)\
                           .join(model.Person)\
                           .order_by(model.Person.first_name,
                                     model.Person.last_name)
        rows = []

        def add_row(customer, i):
            person = customer.account_holder
            address = customer.address or person.address
            rows.append({
                'first_name': person.first_name,
                'last_name': person.last_name,
                'street': address.street if address else None,
                'street2': address.street2 if address else None,
                'city': address.city if address else None,
                'state': address.state if address else None,
                'zipcode': address.zipcode if address else None,
                'address_invalid': address.invalid if address else None,
            })

        self.progress_loop(add_row, customers, progress,
                           message="Fetching data for report")
        return rows



# -*- coding: utf-8 -*-
"""add product.discontinued

Revision ID: 6ef6db00abe8
Revises: 471299288da9
Create Date: 2017-03-03 12:16:42.683837

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '6ef6db00abe8'
down_revision = u'471299288da9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # product
    op.add_column('product', sa.Column('discontinued', sa.Boolean(), nullable=True))
    product = sa.sql.table('product', sa.sql.column('discontinued'))
    op.execute(product.update().values({'discontinued': False}))
    op.alter_column('product', 'discontinued', nullable=False,
                    existing_type=sa.Boolean())


def downgrade():

    # product
    op.drop_column('product', 'discontinued')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch-related commands
"""

import os
import datetime
import logging
import sys

from rattail.commands.core import Subcommand, date_argument, dict_argument
from rattail.progress import SocketProgress
from rattail.time import localtime, make_utc
from rattail.util import simple_error


log = logging.getLogger(__name__)


class BatchHandlerCommand(Subcommand):
    """
    Base class for commands which invoke a batch handler.

    Note that if such a command fails to run its action, the error
    will be written to STDOUT as a simple message string.  If the full
    traceback is needed, logs must be consulted.  This is done for
    sake of letting the caller capture output for display to the user
    when something goes wrong.
    """

    def add_parser_args(self, parser):
        parser.add_argument('--batch-type', metavar='KEY',
                            help="Type of batch to be dealt with, e.g. 'vendor_catalog'")
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def get_handler(self, args):
        """
        Must return the batch handler to use, per the given ``args``
        object.

        Default logic assumes that ``args`` has a ``batch_type``
        attribute, which determines which batch handler should be
        used.  That is figured out by calling
        :meth:`~rattail.app.AppHandler.get_batch_handler()` on the app
        handler.

        :param args: Reference to the
           :class:`python:argparse.Namespace` instance, which was the
           result of parsing the command line args.

        :returns: Must return the batch handler.  This will be an
           instance of some class which derives from
           :class:`~rattail.batch.handlers.BatchHandler`.
        """
        return self.app.get_batch_handler(args.batch_type)

    def run(self, args):
        handler = self.get_handler(args)
        session = self.make_session()
        user = self.get_runas_user(session)

        try:
            success = self.do_stuff(args, handler, session, user)
        except Exception as error:
            log.warning("handler action failed", exc_info=True)
            # nb. only write "simple" error string to stdout, in case
            # caller (e.g. web app) is capturing that for display to
            # user.  admin can consult logs if more info is needed.
            # we do *not* write to stderr, because logging probably
            # is already doing that; we want to avoid mixing them.
            # TODO: this admittedly seems brittle, maybe caller should
            # pass command line flags to control this?
            self.stdout.write(simple_error(error))
            session.rollback()
            session.close()
            sys.exit(42)

        self.finalize_session(session, dry_run=args.dry_run,
                              success=success)

    def do_stuff(self, args, handler, session, user):
        raise NotImplementedError


class MakeBatch(BatchHandlerCommand):
    """
    Make a new batch, from a data file
    """
    name = 'make-batch'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        super(MakeBatch, self).add_parser_args(parser)

        parser.add_argument('--input-file', metavar='PATH', required=True,
                            help="Path to single input file, to be used as data "
                            "source for the new batch.  (File format will vary "
                            "depending on batch type.)")

        parser.add_argument('--kwargs', type=dict_argument, default={},
                            help="Optional JSON-encoded string containing extra "
                            "keyword arguments to be passed to the handler's batch "
                            "creation logic.")

    def do_stuff(self, args, handler, session, user):
        """
        This will create a new batch of the specified type, then populate it
        with the given data file.
        """
        if not os.path.exists(args.input_file):
            raise RuntimeError("input file path does not exist: {}".format(args.input_file))

        kwargs = dict(args.kwargs)
        delete_if_empty = kwargs.pop('delete_if_empty', False)
        auto_execute_allowed = kwargs.pop('auto_execute_allowed', True)

        batch = handler.make_batch(session, created_by=user, **kwargs)
        handler.set_input_file(batch, args.input_file)
        handler.do_populate(batch, user)

        if delete_if_empty:
            session.flush()
            if not batch.data_rows:
                log.debug("auto-deleting empty '%s' batch: %s", handler.batch_key, batch)
                handler.do_delete(batch, dry_run=args.dry_run,
                                  progress=self.progress)
                batch = None

        if batch and auto_execute_allowed and handler.auto_executable(batch):
            handler.execute(batch, user=user)
            batch.executed = make_utc()
            batch.executed_by = user

        return True


class BatchAction(BatchHandlerCommand):
    """
    Base class for commands which invoke a handler to act on a single batch.
    """

    def add_parser_args(self, parser):
        super(BatchAction, self).add_parser_args(parser)
        parser.add_argument('batch_uuid',
                            help="UUID of the batch to be populated.")

    def do_stuff(self, args, handler, session, user):
        """
        This will invoke some action on the batch.
        """
        batch = session.get(handler.batch_model_class, args.batch_uuid)
        if not batch:
            raise RuntimeError("Batch of type '{}' not found: {}".format(args.batch_type, args.batch_uuid))

        return self.action(args, handler, batch, user)


class AutoReceiveBatch(BatchAction):
    """
    Auto-receive all items in a receiving batch
    """
    name = 'auto-receive'
    description = __doc__.strip()

    def action(self, args, handler, batch, user):
        return handler.auto_receive_all_items(batch, progress=self.progress)


class PopulateBatch(BatchAction):
    """
    Populate initial data for a batch
    """
    name = 'populate-batch'
    description = __doc__.strip()

    def action(self, args, handler, batch, user):
        return handler.do_populate(batch, user, progress=self.progress)


class RefreshBatch(BatchAction):
    """
    Refresh data for a batch
    """
    name = 'refresh-batch'
    description = __doc__.strip()

    def action(self, args, handler, batch, user):
        return handler.do_refresh(batch, user, progress=self.progress)


class ExecuteBatch(BatchAction):
    """
    Execute a batch
    """
    name = 'execute-batch'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        super(ExecuteBatch, self).add_parser_args(parser)

        parser.add_argument('--kwargs', type=dict_argument, default={},
                            help="Optional JSON-encoded string containing extra "
                            "keyword arguments to be passed to the handler's batch "
                            "execution function.")

    def action(self, args, handler, batch, user):
        return handler.do_execute(batch, user, progress=self.progress, **args.kwargs)


class PurgeBatches(BatchHandlerCommand):
    """
    Purge old batches from the database
    """
    name = 'purge-batches'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        super(PurgeBatches, self).add_parser_args(parser)

        parser.add_argument('--before', type=date_argument, metavar='DATE',
                            help="Purge all batches executed prior to this date.  If not "
                            "specified, will use --before-days to calculate instead.")

        parser.add_argument('--before-days', type=int, default=90, metavar='DAYS',
                            help="Number of days before the current date, to be used "
                            "as the cutoff date if --before is not specified.  Default "
                            "is 90 days before current date.")

        parser.add_argument('--list-types', action='store_true',
                            help="If set, list available batch types instead of trying "
                            "to purge anything.")


    def run(self, args):
        if args.list_types:
            self.list_types()
            return

        handler = self.get_handler(args)
        session = self.make_session()

        kwargs = {
            'dry_run': args.dry_run,
            'progress': self.progress,
        }
        if args.before:
            before = datetime.datetime.combine(args.before, datetime.time(0))
            before = localtime(self.config, before)
            kwargs['before'] = before
        else:
            kwargs['before_days'] = args.before_days

        purged = handler.purge_batches(session, **kwargs)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

    def list_types(self):
        from rattail.batch.handlers import get_batch_types

        keys = get_batch_types(self.config)
        for key in keys:
            self.stdout.write("{}\n".format(key))



# -*- coding: utf-8 -*-
"""add label_batch.filename

Revision ID: 19da5ebe6fb5
Revises: 36ffa8b83782
Create Date: 2016-11-19 15:59:36.893529

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '19da5ebe6fb5'
down_revision = u'36ffa8b83782'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # label_batch
    op.add_column('label_batch', sa.Column('filename', sa.String(length=255), nullable=True))


def downgrade():

    # label_batch
    op.drop_column('label_batch', 'filename')



# -*- coding: utf-8; -*-
"""add membership_type

Revision ID: 2ff00a7cb734
Revises: 4747a017b8f9
Create Date: 2023-06-06 12:11:21.922167

"""

# revision identifiers, used by Alembic.
revision = '2ff00a7cb734'
down_revision = '4747a017b8f9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # membership_type
    op.create_table('membership_type',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.String(length=20), nullable=True),
                    sa.Column('number', sa.Integer(), nullable=True),
                    sa.Column('name', sa.String(length=100), nullable=True),
                    sa.PrimaryKeyConstraint('uuid')
                    )
    op.create_table('membership_type_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('id', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('number', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_membership_type_version_end_transaction_id'), 'membership_type_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_membership_type_version_operation_type'), 'membership_type_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_membership_type_version_transaction_id'), 'membership_type_version', ['transaction_id'], unique=False)

    # member
    op.add_column('member', sa.Column('membership_type_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('member_fk_membership_type', 'member', 'membership_type', ['membership_type_uuid'], ['uuid'])
    op.add_column('member_version', sa.Column('membership_type_uuid', sa.String(length=32), autoincrement=False, nullable=True))


def downgrade():

    # member
    op.drop_column('member_version', 'membership_type_uuid')
    op.drop_constraint('member_fk_membership_type', 'member', type_='foreignkey')
    op.drop_column('member', 'membership_type_uuid')

    # membership_type
    op.drop_index(op.f('ix_membership_type_version_transaction_id'), table_name='membership_type_version')
    op.drop_index(op.f('ix_membership_type_version_operation_type'), table_name='membership_type_version')
    op.drop_index(op.f('ix_membership_type_version_end_transaction_id'), table_name='membership_type_version')
    op.drop_table('membership_type_version')
    op.drop_table('membership_type')



# -*- coding: utf-8; -*-
"""normalize catalog percents

Revision ID: dc28b97c33ff
Revises: ffd65c52b0fd
Create Date: 2022-11-28 11:31:21.447456

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'dc28b97c33ff'
down_revision = 'ffd65c52b0fd'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


batch_vendorcatalog_row = sa.sql.table('batch_vendorcatalog_row',
                                       sa.sql.column('unit_cost_diff_percent'),
                                       sa.sql.column('discount_percent'))


def convert_all(convert):

    # batch_vendorcatalog_row
    op.execute(batch_vendorcatalog_row.update()\
               .where(batch_vendorcatalog_row.c.unit_cost_diff_percent != None)\
               .values({'unit_cost_diff_percent': convert(batch_vendorcatalog_row.c.unit_cost_diff_percent)}))
    op.execute(batch_vendorcatalog_row.update()\
               .where(batch_vendorcatalog_row.c.discount_percent != None)\
               .values({'discount_percent': convert(batch_vendorcatalog_row.c.discount_percent)}))


def upgrade():

    # input values are 0.0 - 1.0 range, output is 0 - 100
    convert_all(lambda val: val * 100)


def downgrade():

    # input values are 0 - 100 range, output is 0.0 - 1.0
    convert_all(lambda val: val / 100.0)



# -*- coding: utf-8; -*-
"""add vendor invoice cost diff

Revision ID: 3ed719791439
Revises: 8e8c71d67fb8
Create Date: 2019-10-09 15:40:50.667401

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '3ed719791439'
down_revision = '8e8c71d67fb8'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor_invoice_row
    op.add_column('vendor_invoice_row', sa.Column('unit_cost_diff', sa.Numeric(precision=9, scale=5), nullable=True))


def downgrade():

    # vendor_invoice_row
    op.drop_column('vendor_invoice_row', 'unit_cost_diff')
